package dev.pfaff.altargui.util;

import dev.pfaff.altargui.function.FallibleRunnable;
import org.jspecify.annotations.Nullable;

import java.util.concurrent.Callable;
import java.util.concurrent.CompletionException;

public final class ThreadFuture<V> implements BasicFuture<V> {
	private static final Thread DEAD_THREAD = Thread.startVirtualThread(() -> {});

	private final Thread thread;
	private final ResultSlot<V> resultSlot;

	private ThreadFuture(Thread thread, ResultSlot<V> slot) {
		this.thread = thread;
		this.resultSlot = slot;
	}

	public static ThreadFuture<Void> spawnVirtualTask(FallibleRunnable runnable) {
		var slot = new ResultSlot<Void>();
		return new ThreadFuture<>(Thread.startVirtualThread(() -> {
			try {
				runnable.run();
			} catch (Throwable e) {
				slot.throwable = e;
			}
		}), slot);
	}

	public static <V> ThreadFuture<V> spawnVirtualTask(Callable<V> callable) {
		var slot = new ResultSlot<V>();
		return new ThreadFuture<>(Thread.startVirtualThread(() -> {
			try {
				slot.value = callable.call();
			} catch (Throwable e) {
				slot.throwable = e;
			}
		}), slot);
	}

	public V join() {
		try {
			this.thread.join();
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
		if (this.resultSlot.throwable != null) throw new CompletionException(this.resultSlot.throwable);
		return this.resultSlot.value;
	}

	public boolean isAlive() {
		return this.thread.isAlive();
	}

	private static final class ResultSlot<V> {
		private V value;
		@Nullable
		private Throwable throwable;

		public ResultSlot() {
		}

		public ResultSlot(V value) {
			this.value = value;
		}

		public ResultSlot(Throwable throwable) {
			this.throwable = throwable;
		}
	}
}
