package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.element.CustomLayoutSingleChildWidgetInstance;
import org.jetbrains.annotations.NotNull;

/**
 * A widget that calls the provided {@code transformer} every layout.
 */
public record CustomTransformBuilder(Widget child, Transformer transformer) implements Widget {
	@Override
	public @NotNull WidgetInstance<CustomTransformBuilder> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	public static final class Transform {
		public float x;
		public float y;

		private Transform(float x, float y) {
			this.x = x;
			this.y = y;
		}
	}

	@FunctionalInterface
	public interface Transformer {
		void transform(Transform transform,
					   float computedWidth, float computedHeight,
					   float availableWidth, float availableHeight);
	}

	private static final class State extends CustomLayoutSingleChildWidgetInstance<CustomTransformBuilder> {
		private Transformer transformer;
		private float transformedX;
		private float transformedY;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, CustomTransformBuilder widget, boolean init) {
			setChild(stateContext, widget.child);
			if (this.transformer != widget.transformer) {
				this.transformer = widget.transformer;
				this.markNeedsLayout();
			}
		}

		@Override
		protected void computeStencilLayout(StencilLayoutContext context) {
			context.layout(this, getChildElement(), (e, ctx) -> {
				e.copyAvailableSizeFrom(this);
				e.stencilLayout(ctx);
				var transform = new Transform(e.allocatedX(), e.allocatedY());
				this.transformer.transform(transform,
						e.computedWidth(),
						e.computedHeight(),
						e.allocatedWidth(),
						e.allocatedHeight());
				this.transformedX = transform.x;
				this.transformedY = transform.y;
				// TODO: is this really the ideal behaviour?
				copyComputedSizeFrom(e);
			});
		}

		@Override
		protected void computeTraceLayout() {
			var e = getChildElement();
			// TODO: should we allow overflow?
			e.setPosition(this.transformedX, this.transformedY);
			e.traceLayout();
		}
	}
}
