package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.*;
import dev.pfaff.altargui.handlers.MouseScrollHandler;
import dev.pfaff.altargui.element.CustomLayoutSingleChildWidgetInstance;
import dev.pfaff.altargui.inspector.DebugLayer;
import org.jetbrains.annotations.NotNull;

public record ScrollWidget(Widget child) implements Widget {
	public static final DebugLayer DEBUG_LAYER_PRE_UNSCISSORED = new DebugLayer(ScrollWidget.class, "pre_unscissored");
	public static final DebugLayer DEBUG_LAYER_PRE_SCISSORED = new DebugLayer(ScrollWidget.class, "pre_scissored");
	public static final DebugLayer DEBUG_LAYER_POST_UNSCISSORED = new DebugLayer(ScrollWidget.class, "post_unscissored");

	@Override
	public @NotNull WidgetInstance instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends CustomLayoutSingleChildWidgetInstance<ScrollWidget> implements MouseScrollHandler {
		/**
		 * The scroll offset.
		 */
		private double offset;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, ScrollWidget widget, boolean init) {
			setChild(stateContext, widget.child);
		}

		@Override
		protected void computeStencilLayout(StencilLayoutContext context) {
			final var allocWidth = this.allocatedWidth();
			final var allocHeight = this.allocatedHeight();

			var childElement = getChildElement();
			context.layout(this, childElement, (e, ctx) -> {
				e.setAvailableSize(allocWidth, allocHeight);
				e.stencilLayout(ctx);
			});

			float height = childElement.computedHeight();
			if (offset < 0.0 && height < allocHeight) {
				offset = Math.min(0, allocHeight - height);
			}

			setComputedSize(childElement.computedWidth(), Math.min(height, allocHeight));
		}

		@Override
		protected void computeTraceLayout() {
			var y = (float) (this.allocatedY() + offset);

			var childElement = getChildElement();
			childElement.setPosition(allocatedX(), y);
			childElement.traceLayout();
		}

		@Override
		protected void computeMount(MountContext context) {
			context.pushScissor(this);
			super.computeMount(context);
			context.popScissor();
		}

		@Override
		public void uSpecMount(MountContext context) {
			super.uSpecMount(context);
		}

		@Override
		public boolean mouseScrolled(double mouseX, double mouseY, double delta) {
			if (isInside(mouseX, mouseY)) {
				offset = Math.min(0, offset + MouseScrollHandler.applyStandardDeltaTransform(delta));
				markNeedsLayout();
				return true;
			}
			return false;
		}
	}
}
